/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.common;

import java.util.*;

import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.GridPoint;

/**
 * Really simple "blob detection". When given a grid, an initial point and a function
 * which defines whether points can be used, it will return a set of points which
 * are part of a connected "blob" starting from the given point.
 * 
 * The membership function is essentially a mask, so you can set up whatever conditions 
 * you need for your blobs. Examples would be cells with certain land uses (membership function
 * returns true if it is an allowed land use) or with a certain number of people (returns
 * true if numPeople > threshold) etc.
 * 
 * It uses the Von Neumann neighbourhood (or 4-connected blobs in image processing terms).
 * 
 * An optional halting condition function allows the search to terminate early if certain
 * conditions are met.
 * @author dmrust
 *
 */
public class ConnectedPoints
{
	int[] size;
	int[][] offsets = { {-1, 0}, {1,0}, {0,-1}, {0,1} };
	MembershipFunction membership;
	HaltingCondition halt;
	
	
	/**
	 * Constructs a new ConnectedPoints, with the given membership function and halting condition,
	 * assuming a grid of size size[0] by size[1]. HaltingCondition may be null, but MembershipFunction must
	 * not be null.
	 * @param member the boolean function for whether a point is a member of the set
	 * @param halting a condition for when the search should be stopped
	 * @param size an int array defining the size of the grid
	 */
	public ConnectedPoints( MembershipFunction member, HaltingCondition halting, int...size )
	{
		this.size = size;
		this.membership = member;
		this.halt = halting;
	}
	
	public Set<GridPoint> getConnectedPoints( GridPoint initial )
	{
		HashSet<GridPoint> g = new HashSet<GridPoint>();
		g.add( initial );
		return getConnectedPoints( g );
	}
	
	/**
	 * Returns the set of connected points starting from the inital point.
	 * @param initial
	 * @return
	 */
	public Set<GridPoint> getConnectedPoints( Set<GridPoint> initial )
	{
		return getConnectedPoints( initial, false );
	}
	
	/**
	 * Returns the set of connected points starting from the initial set.
	 * 
	 * The initial sanity check, which makes sure that the given set of points
	 * are members can be skipped (e.g. for times when the membership function
	 * only returns true for things which are not already a member)
	 * @param initial
	 * @param skipCheck
	 * @return
	 */
	public Set<GridPoint> getConnectedPoints( Set<GridPoint> initial, boolean skipCheck )
	{
		Set<GridPoint> points = new HashSet<GridPoint>();
		points.addAll( initial );
		
		//Quick sanity check - return nothing if initial point is not a member
		if( ! skipCheck )
		{
			Iterator<GridPoint> it = points.iterator();
			while( it.hasNext() )
			{
				if( ! membership.isMember( it.next() ))
					it.remove();
			}
			if( points.size() == 0 ) return points;
		}
		
		System.out.println( "Initial membership: " + points.size());
		
		//Setup current blob, seen pixels, queue of pixels to look at
		Set<GridPoint> seen = new HashSet<GridPoint>();
		Queue<GridPoint> queue = new LinkedList<GridPoint>();
		queue.addAll( initial );
		seen.addAll( initial );
		
		Set<GridPoint> neighbours = new HashSet<GridPoint>();
		//While we have a queue of cells to check
		while( queue.size() > 0 )
		{
			neighbours.clear();
			//Get the neighbours
			GridPoint current = queue.poll();
			
			//Add the neighbours which we havent seen, and which satisfy the membership function
			//Make sure we do them in a random order so there's no spatial bias
			int offset = RandomHelper.nextIntFromTo( 0, 3 );
			System.out.println( "Expanding around " + current );
			for( int i = 0; i < offsets.length; i++ )
			{
				int[] os = offsets[(i+offset)%offsets.length];
				int x = current.getX() + os[0];
				int y = current.getY() + os[1];
				if( x >= 0 && x < size[0] && y >= 0 && y < size[1] )
				{
					neighbours.add( new GridPoint( x, y ) );
					System.out.println( "Adding neighbour: " + x + "," + y );
				}
			}
			
			for( GridPoint g : neighbours )
			{
				if( ! seen.contains( g ) )
				{
					seen.add( g );
					if( membership.isMember( g ) )
					{
						points.add( g );
						queue.add( g );
						//checking the halting condition as we go
						if( halt != null && halt.isDone( queue, seen, points ))
							return points;
					}
					
				}
			}
		}
		return points;
	}
	

	/**
	 * A boolean function, which defines whether a point is allowable
	 * @author dmrust
	 *
	 */
	public interface MembershipFunction {
		public boolean isMember( GridPoint g );
	}
	
	/**
	 * Returns a boolean to indicate if the search should terminate early
	 * @author dmrust
	 *
	 */
	public interface HaltingCondition {
		public boolean isDone( Queue<GridPoint> queue, Set<GridPoint> seen, Set<GridPoint> points );
	}
	
	/**
	 * Simple halting condition to stop when the blob is a certain size.
	 * @param max
	 * @return
	 */
	public static HaltingCondition maxPointsCondition( final int max )
	{
		return new HaltingCondition()
		{
			public boolean isDone( Queue<GridPoint> queue, Set<GridPoint> seen, Set<GridPoint> points )
			{
				if( points.size() >= max ) return true;
				return false;
			}
		};
	}

	public void setMembership( MembershipFunction membership )
	{
		this.membership = membership;
	}

	public void setHalt( HaltingCondition halt )
	{
		this.halt = halt;
	}

	public void setSize( int[] size )
	{
		this.size = size;
	}
}
