/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.common;

import static abmland.common.LogExceptionHandler.handleException;
import static abmland.common.Properties.ASSAYAREA;
import static abmland.common.Utils.evalForeignMethod;
import static org.apache.log4j.Logger.getLogger;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;

import java.io.File;
import java.net.URL;
import java.util.Map;
import java.util.Map.Entry;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import repast.simphony.context.AbstractContext;
import repast.simphony.context.Context;
import repast.simphony.context.DefaultContext;
import repast.simphony.context.space.grid.GridFactoryFinder;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.DefaultScheduleRegistryFactory;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.environment.ScheduleRegistry;
import repast.simphony.engine.schedule.DefaultSchedulableActionFactory;
import repast.simphony.engine.schedule.DefaultScheduleFactory;
import repast.simphony.engine.schedule.ISchedulableAction;
import repast.simphony.engine.schedule.ISchedulableActionFactory;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.watcher.Watch;
import repast.simphony.engine.watcher.WatchItems;
import repast.simphony.engine.watcher.WatcherTriggerSchedule;
import repast.simphony.parameter.IllegalParameterException;
import repast.simphony.parameter.ListTokenizer;
import repast.simphony.parameter.ParameterFormatException;
import repast.simphony.parameter.Parameters;
import repast.simphony.parameter.ParametersCreator;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridBuilderParameters;
import repast.simphony.space.grid.GridDimensions;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.space.grid.RandomGridAdder;
import repast.simphony.space.grid.StrictBorders;
import repast.simphony.space.projection.Projection;
import repast.simphony.util.collections.IndexedIterable;
import repast.simphony.valueLayer.ValueLayer;
import abmland.data.CLCover;

public class RepastUtils {

	private final static Logger log = getLogger(RepastUtils.class);

	private final static ClassLoader classLoader = RepastUtils.class
			.getClassLoader();

	public synchronized void retrieveModelFromChildContext(
			Context<? extends Object> mainContext) {
		// The model-impl. are assorted directly under the main context,
		// so retrieve them...
		for (Context<? extends Object> childcontext : mainContext
				.getSubContexts()) {
			for (Object model : childcontext) {
				if (Pattern.matches("\\p{Alnum}*Model", model.getClass()
						.getSimpleName())) {

					log
							.debug("He, I found "
									+ model.getClass().getSimpleName());
					// Call a {interface} function in the model
					// Object hashCode =
					evalForeignMethod(model, "sayHello");
					log.debug(model.getClass().getSimpleName() + " hashcode "
							+ evalForeignMethod(model, "hashCode"));
				}
			}
		}
	}

	/**
	 * Using the action's priorities as quasi-ID, so give distinct annotation
	 * attribute values in the agent definitions.
	 *
	 * @param annotatedObj
	 */
	public static void getMethodsForSchedulingAsActionList(Object annotatedObj) {

		ISchedulableActionFactory f = new DefaultSchedulableActionFactory();
		for (ISchedulableAction schedulable : f.createActions(annotatedObj)) {

			log.info("Action " + schedulable.getPriority() + " added at index "
					+ schedulable.getOrderIndex() + " to the schedule.");
		}
	}

	/**
	 * @deprecated TODO Daniel Figure out how to integrate other models. query
	 *             all models in your sub-contexts
	 *
	 *             FIXME If the below doesn't work with this repasts version,
	 *             split in up to no-of sub-models
	 *
	 *             watch-annotated boolean-typed syntactical comma-separated
	 *             queries, each looking for the finished value in their model
	 *             and finally AND-aggregating those to the finish-result used
	 *             by the triggerCompute method-annotation... {{{
	 */
	@WatchItems( //
	watches = { @Watch( //
			watcheeClassName = "ResidentModel", //
			watcheeFieldNames = "getData_of_ResidentModel_finished", //
			triggerCondition = "watchee.getData_of_ResidentModel_finished", //
			whenToTrigger = WatcherTriggerSchedule.IMMEDIATE), //
			@Watch( //
			watcheeClassName = "", //
			watcheeFieldNames = "", //
			whenToTrigger = WatcherTriggerSchedule.IMMEDIATE) })
	public void multipleWatcherQuery() {
		;
	}

	public static <T> ValueLayer CreateValueLayer(String layerName,
			Context<T> context, int... dims) {

		return null;
	}

	public static <T> Grid<T> CreateMultiOccupancy2DGrid(String gridName,
			Context<T> context, int... dims) {
		assertEquals(dims.length, 2);
		GridFactoryFinder.createGridFactory(null).createGrid(
				gridName,
				context,
				new GridBuilderParameters<T>(new StrictBorders(),
						new RandomGridAdder<T>(), true, dims[0], dims[1]));

		Grid<T> grid = (Grid<T>) context.getProjection(gridName);

		return grid;
	}

	public static void SetMasterContext(Context<?> masterContext) {
		RunState.getInstance().setMasterContext(masterContext);
	}

	public static <T> Grid<T> CreateSingleOccupancy2DGrid(String gridName,
			Context<T> context, int... dims) {

		// --------------- Reworked ---------------

		assertEquals(dims.length, 2);
		GridFactoryFinder.createGridFactory(null).createGrid(
				gridName,
				context,
				new GridBuilderParameters<T>(new StrictBorders(),
						new RandomGridAdder<T>(), false, dims[0], dims[1]));

		Grid<T> grid = (Grid<T>) context.getProjection(gridName);

		return grid;

		// --------------- Original ---------------
		/*
		 * assertEquals(dims.length, 2);
		 *
		 * GridBuilderParameters<T> params = GridBuilderParameters
		 * .singleOccupancy2D(new RandomGridAdder<T>(), new StrictBorders(),
		 * dims[0], dims[1]);
		 *
		 * // borrowed here: //
		 * repast.simphony.dataLoader.engine.GridProjectionBuilder
		 * .GridBuilder.build(Context) GridFactory factory =
		 * GridFactoryFinder.createGridFactory(null);
		 *
		 * return factory.createGrid(gridName, context, params);
		 */
	}

	/**
	 * @param <I>
	 * @param contained
	 * @return the context which directly contains object o, starting from the
	 *         master context
	 * @see AbstractContext#findParent(Object)
	 */
	// public static Context<?> findParent(Object contained) {
	// // un-optimized
	// Context<?> context = null;
	// // see repast.simphony.essentials.RETest.setUp()
	// try {
	// RunState instance = RunState.getInstance();
	// instance.hashCode();
	// context = instance.getMasterContext().findParent(contained);
	// } catch (NullPointerException e) {
	// log.error("No RunState instance found.", e);
	// }
	// return context;
	// }

	@SuppressWarnings("unchecked")
	public static <I> Context<I> findParent(I contained) {
		// un-optimized
		Context<I> context = null;
		// see repast.simphony.essentials.RETest.setUp()
		context = getMasterContext().findParent(contained);
		return context;
	}

	/**
	 * @param <I>
	 * @param contained
	 * @return the first context found, which contains an object of this type,
	 *         the search beginning at master context level.
	 */
	@SuppressWarnings("unchecked")
	public static <I> Context<I> findParent(Class<I> contained) {
		// un-optimized
		IndexedIterable<I> objectsOfTypeI = null;
		// see repast.simphony.essentials.RETest.setUp()
		objectsOfTypeI = (IndexedIterable<I>) getMasterContext().getObjects(
				contained);
		for (I object : objectsOfTypeI) {
			return findParent(object);
		}
		return null;
	}

	public static GridDimensions GetSimulationDimensions(/** String gridName */
	) {
		GridDimensions dimensions = null;
		// see repast.simphony.essentials.RETest.setUp()
		Projection<?> projection = null;
		projection = getMasterContext().getProjection(ASSAYAREA);
		if (null == projection) {
			try {
				assertNotNull(getMasterContext().getSubContexts());
				for (Object subcontext : getMasterContext().getSubContexts()) {
					projection = ((Context<?>) subcontext)
							.getProjection(ASSAYAREA);
					if (null != projection) {
						dimensions = ((Grid<?>) projection).getDimensions();

						break;
					}
				}
				try {
					assertNotNull(projection);
					try {
						assertNotNull(dimensions);
					} catch (Throwable no_dimensions) {
						log.error("", no_dimensions);
						System.exit(-1);
					}
				} catch (Throwable no_assay_area) {
					log
							.error(
									"Likely there is no AssayArea projection defined. ",
									no_assay_area);
					System.exit(-1);
				}
			} catch (Throwable no_subctx) {
				log.error("", no_subctx);
				System.exit(-1);
			}
		} else {
			try {
				assertNotNull(((Grid<?>) projection).getDimensions());
				dimensions = ((Grid<?>) projection).getDimensions();
			} catch (Throwable no_dimensions) {
				log.error("", no_dimensions);
				System.exit(-1);
			}
		}
		return dimensions;
	}

	private static RunEnvironment tryRunEnvironment() {
		RunEnvironment _ = RunEnvironment.getInstance();
		if (null == _)
			throw new NullPointerException(
					"There is no RunEnvironment instance.");
		return _;
	}

	private static Parameters tryParameters(RunEnvironment runEnvironment) {
		Parameters _ = runEnvironment.getParameters();
		if (null == _)
			throw new NullPointerException("Parameters not read.");
		return _;
	}

	private static Integer[] tryDims(Parameters parameters) throws ParameterFormatException {
		Integer[] _;
		try {
			_ = ListTokenizer.parseIntValues(parameters
					.getValueAsString(ASSAYAREA), ASSAYAREA);
		} catch (IllegalParameterException e) {
			throw e;
		}
		return _;
	}

	/**
	 * Assumes there is a extended_params.xml file in the .rs folder given as
	 * parameter (giving as parameter is also assumed) and in that file there is
	 * a parameter named ${ASSAYAREA} set.
	 *
	 * @return
	 * @throws Exception
	 * @throws ParameterFormatException
	 */
	public static int[] GetSimulationDimensionsIntArray() {
		// Get Parameters from RunEnvironment
		log.trace(null);
		int[] _ = null;
		RunEnvironment runEnvironment = tryRunEnvironment();

		Parameters params = tryParameters(runEnvironment);

		Integer[] assayareadims = null;
		try {
			assayareadims = tryDims(params);
		} catch (ParameterFormatException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		 _ = new int[assayareadims.length];
		 short pos = 0;
		 for (int i : assayareadims) {
		 _[pos] = i;
		 }
		return _;
	}

	public static int[] GetSimulationDimensionsIntArrayOLD() {
		GridDimensions dim = GetSimulationDimensions();
		return dim.toIntArray(null);
	}

	public static double getCurrentSimulationTick() {
		// un-optimized
		double tick = 0;
		tick = getModelSchedule().getTickCount();
		log.debug("tick = " + tick);
		return tick;
	}

	public static ISchedule getModelSchedule() {
		// un-optimized
		// IScheduleFactory
		// ISchedulableActionFactory
		// see repast.simphony.essentials.RETest.setUp()
		int EXIT = -1;
		ISchedule iSchedule = null;
		ScheduleRegistry scheduleRegistry = null;
		RunState runstate = RunState.getInstance();
		try {
			assertNotNull(runstate);
			try {
				assertNotNull(runstate.getScheduleRegistry());
				scheduleRegistry = runstate.getScheduleRegistry();
				try {
					assertNotNull(scheduleRegistry.getModelSchedule());
					iSchedule = scheduleRegistry.getModelSchedule();
				} catch (Throwable no_ischedule) {
					log.error("ISchedule not found. ", no_ischedule);
					// System.exit(EXIT);
				}
			} catch (Throwable no_scheduleregistry) {
				log.error("ScheduleRegistry not found. ", no_scheduleregistry);
				// System.exit(EXIT);
			}
		} catch (Throwable no_runstate) {
			log.error("RunState not initialized. ", no_runstate);
			// System.exit(EXIT);
		}
		if (iSchedule == null) {
			RunEnvironment runenv = RunEnvironment.getInstance();
			try {
				assertNotNull(runenv);
				try {
					assertNotNull(runenv.getCurrentSchedule());
					iSchedule = runenv.getCurrentSchedule();
				} catch (Throwable no_ischedule) {
					log.error("ISchedule not found. ", no_ischedule);
					// System.exit(EXIT);
				}
			} catch (Throwable no_runenvironment) {
				log.error("No RunEnvironment initialized. ", no_runenvironment);
				// System.exit(EXIT);
			}
		}
		return iSchedule;
	}

	public static Context<Object> getMasterContext() {
		Context<Object> mContext = null;
		RunState runstate = RunState.getInstance();
		try {
			assertNotNull(runstate);
			try {
				assertNotNull(runstate.getMasterContext());
				mContext = runstate.getMasterContext();
			} catch (Throwable no_masterCtx) {
				log.error("No MasterContext associated with RunState. ",
						no_masterCtx);
				System.exit(0);
			}
		} catch (Throwable no_runstate) {
			log.error("No RunState initialized. ", no_runstate);
			System.exit(0);
		}
		return mContext;
	}

	/**
	 * TODO Put in a public class for all models.
	 *
	 * @param <T>
	 * @param clazz
	 * @return The objects found - usually in a context as an indexed iterable
	 *         collection.
	 */
	public static <T> IndexedIterable<T> findObjects(Class<T> clazz) {
		String clazzName = clazz.getName();
		IndexedIterable<T> iterable = null;
		try {
			iterable = (IndexedIterable<T>) getMasterContext().getObjects(
					Class.forName(clazzName));
		} catch (ClassNotFoundException e) {
			log.error(clazzName + " typed class not locatable.", e);
		}
		return iterable;
	}

	public static <T> T getModelController(Class<T> clazz) {
		T type = null;
		try {
			for (T obj : findObjects(clazz)) {
				type = obj;
			}
		} catch (NullPointerException e) {
			log.error("No objects of type " + type + " found.", e);
		}
		return type;
	}

	/**
	 * TODO Maybe unused... because Parameter setting not integrated
	 *
	 * This method should prepare a simulation with
	 * <ul>
	 * <li>RunState initialized</li>
	 * <li>Scheduling enabled/prepared</li>
	 * <li>RunEnvironment initialized</li>
	 * </ul>
	 */
	public static void ResetDefaultSimulationRun() {
		RunState.init(null, null, null);
		ScheduleRegistry scheduleRegistry = new DefaultScheduleRegistryFactory()
				.createScheduleRegistry();
		RunState.getInstance().setScheduleRegistry(scheduleRegistry);
		ISchedule schedule = new DefaultScheduleFactory().createSchedule();
		RunState.getInstance().getScheduleRegistry().setModelSchedule(schedule);
		// TODO is this RunEnvironment needed in all default cases?
		RunEnvironment.init(schedule, null, null, false);
	}

	private class MasterContext implements ContextBuilder<Object> {

		/** {@inheritDoc} */
		public Context<Object> build(Context<Object> context) {
			context.addSubContext(new IntegerContext());
			return null;
		}

		class IntegerContext extends DefaultContext<Integer> {
			public IntegerContext() {
				super("IntegerContext");
			}
		}
	}

	public static String gridToString(Grid<? extends Object> grid) {
		String ret = "";
		for (int j = (int) grid.getDimensions().getHeight() - 1; j >= 0; j--) {
			for (int i = 0; i < grid.getDimensions().getWidth(); i++)
				ret += grid.getObjectAt(i, j) + "\t";
			ret += "\n";
		}
		return ret;
	}

	public static String layerToString(ValueLayer layer) {
		String ret = "";
		for (int j = (int) layer.getDimensions().getHeight() - 1; j >= 0; j--) {
			for (int i = 0; i < layer.getDimensions().getWidth(); i++)
				ret += layer.get(i, j) + "\t";
			ret += "\n";
		}
		return ret;
	}

	public static String planToString(Map<GridPoint, CLCover> plan) {
		String p = "(";
		for (Entry<GridPoint, CLCover> g : plan.entrySet())
			p += g;
		p += ") => " + plan.values().iterator().next().toString();
		return p;
	}

	/**
	 * Returns a Simphony runner which can i. e. load parameters from
	 * extended_params.xml in a .rs folder, evaluate the model.score file and
	 * act upon etc. In short it should behave like the Runner started when
	 * using the launch files provided by the Simphony distribution.
	 *
	 * @param rsfolder
	 * @return
	 */
	public static FileInitializableScheduleRunner GetIntializableRunner(
			String rsfolder) {

		FileInitializableScheduleRunner runner = new DummyScheduleRunner();

		/**
		 * {{{ The folder with the model.score is expected on the test class
		 * path of the project
		 */
		URL url = classLoader.getResource(rsfolder);

		assertNotNull(url);

		try {
			File scenario = new File(url.toURI());
			try {
				runner.load(scenario);
			} catch (Throwable t) {
				handleException(RepastUtils.class, t);
			}
		} catch (Throwable t) {
			handleException(RepastUtils.class, t);
		}
		/**
		 * }}} ... folder with the model.score...
		 */

		return runner;
	}

	/**
	 * Adds the specified parameter to the current instance of the
	 * RunEnvironment.
	 *
	 * @param name
	 * @param type
	 * @param value
	 * @param isReadOnly
	 */
	public static void AddParameterToRunEnvironment(String name, Class<?> type,
			Object value, boolean isReadOnly) {

		try {
			assertNotNull(RunEnvironment.getInstance());
		} catch (Throwable t) {
			log.error("Please do RunEnvironment.init first.", t);
			fail();
		}

		/**
		 * {{{ We want to add a Dimensions parameter...
		 */
		ParametersCreator creator = new ParametersCreator();

		creator.addParameter(name, type, value, isReadOnly);

		// Reuse existing parameters...
		for (String parameter : RunEnvironment.getInstance().getParameters()
				.getSchema().parameterNames()) {

			creator.addParameter(parameter, RunEnvironment.getInstance()
					.getParameters().getDisplayName(parameter), RunEnvironment
					.getInstance().getParameters().getClass(), RunEnvironment
					.getInstance().getParameters().getValue(parameter),
					RunEnvironment.getInstance().getParameters().isReadOnly(
							parameter));
		}

		RunEnvironment.getInstance().setParameters(creator.createParameters());
		/**
		 * }}} ... Dimensions parameter...
		 */
	}

	/**
	 * pause for 3 seconds
	 */
	public static void pause3() {
		synchronized (RepastUtils.class) {
			try {
				new Object().wait(3000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
}
