/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.common;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public class TypesArith {

	enum OP {
		MULT, ADD, SUB, DIV
	}

	static int reduce(OP op, int[] array) {
		Integer[] a = new Integer[array.length];
		for (int i = 0; i < array.length; i++) {
			a[i] = array[i];
		}
		return reduce(op, Arrays.asList(a));
	}

	static float reduce(OP op, float[] array) {
		Float[] a = new Float[array.length];
		for (int i = 0; i < array.length; i++) {
			a[i] = array[i];
		}
		return reduce(op, Arrays.asList(a));
	}

	static <E extends Number> E reduce(OP op, E[] array) {
		return reduce(op, Arrays.asList(array));
	}

	static <E extends Number> E reduce(OP op, List<E> list) {
		Number r = 0;
		switch (op) {
		case ADD:
			r = list.remove(0);
			for (E e : list) {
				r = r.floatValue() + e.floatValue();
			}
			break;
		case SUB:
			r = list.remove(0);
			for (E e : list) {
				r = r.floatValue() - e.floatValue();
			}
			break;
		case DIV:
			r = list.remove(0);
			for (E e : list) {
				r = r.floatValue() / e.floatValue();
			}
			break;
		case MULT:
			r = list.remove(0);
			for (E e : list) {
				r = r.floatValue() * e.floatValue();
			}
			break;
		default:
			break;
		}
		return (E) r;
	}

	static int reduceNA(OP op, int[] array) {
		ArrayList<Integer> list = new ArrayList<Integer>();
		Integer[] a = new Integer[array.length];
		for (int i = 0; i < array.length; i++) {
			a[i] = array[i];
		}
		Collections.addAll(list, a);
		return reduceNA(op, list);
	}

	static float reduceNA(OP op, float[] array) {
		ArrayList<Float> list = new ArrayList<Float>();
		Float[] a = new Float[array.length];
		for (int i = 0; i < array.length; i++) {
			a[i] = array[i];
		}
		Collections.addAll(list, a);
		return reduceNA(op, list);
	}

	/**
	 * Reduce ignoring absorbing element (i.e. for dimension specifications with
	 * a zero)
	 *
	 * @param <E>
	 * @param op
	 * @param list
	 * @return the value as result of MULT reductions or the normal reductions
	 */
	static <E extends Number> E reduceNA(OP op, ArrayList<E> list) {
		Number r = 0;
		switch (op) {
		case MULT:
			r = list.remove(0);
			if (r instanceof Float) {
				r = (r.floatValue() != 0.0f ? r.floatValue() : 1);
				for (E e : list) {
					r = r.floatValue()
							* (e.floatValue() != 0.0f ? e.floatValue() : 1);
				}
			} else if (r instanceof Integer) {
				r = (r.intValue() != 0 ? r.intValue() : 1);
				for (E e : list) {
					r = r.intValue() * (e.intValue() != 0 ? e.intValue() : 1);
				}
			}
			break;
		default:
			r = reduce(op, list);
			break;
		}
		return (E) r;
	}

}
