/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.mainctrl;

import static abmland.common.CellDumbMock.ABooleanAllCells.requestBooleanAllCells;
import static abmland.common.CellDumbMock.ACLCoverAllCells.requestCLCoverAllCells;
import static abmland.common.CellDumbMock.AFloorSpaceAllCells.requestFloorSpaceAllCells;
import static abmland.common.CellDumbMock.AHouseholdNumberAllCells.requestHouseholdNumberAllCells;
import static abmland.common.LogExceptionHandler.handleException;
import static abmland.common.RepastUtils.getModelController;
import static abmland.common.ScheduleProperties.MAINCTRL_COMPUTE_SCHED_P;
import static abmland.common.ScheduleProperties.MAINCTRL_GETDATA_SCHED_P;
import static abmland.common.ScheduleProperties.MAINCTRL_INIT_SCHED_P;
import static abmland.models.config.RasterIO.raster2clcmap;
import static org.apache.log4j.Logger.getLogger;
import static org.junit.Assert.assertTrue;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.space.grid.GridPoint;
import abmland.data.CLCover;
import abmland.data.Householdtypes;
import abmland.mainctrl.config.ConfManager;
import abmland.mainctrl.config.LanduseConf;
import abmland.mainctrl.config.MiscConf;
import abmland.models.IInfrastructureModel;
import abmland.models.ILU;
import abmland.models.IResidentModel;
import abmland.models.serve.InfrastructureModelServer2;
import abmland.models.serve.ResidentModelServer2;
import abmland.types.IBooleanAllCells;
import abmland.types.ICLCoverAllCells;
import abmland.types.IFloorSpaceAllCells;
import abmland.types.IHouseholdNumber;
import abmland.types.IHouseholdNumberAllCells;
import abmland.types.impl.HouseholdNumber;

import com.moseph.gis.raster.Raster;

/**
 * XXX {p:Paper} At the moment only infrastructure, planner and resident are in.
 *
 * @author dkahlenb
 *
 */
public class LandUseController implements ILU {

	private final Logger log = getLogger(getClass().getName());

	// private Grid<Cell> landUseMap;

	// private Context<Cell> cells = new DefaultContext<Cell>("LandUse");

	private Env env = new Env();

	final class Env {
		// Use this config reference to retrieve things related to
		// land use, like capacity of residential space for a certain CLC or
		// capacity of business space for a certain CLC and things like that.
		LanduseConf config;
		Map<GridPoint, CLCover> PARAM_INITIAL_CLC_MAP;
		Map<GridPoint, Float> clcMap;
		private ICLCoverAllCells exportedLandUse = null;
		IFloorSpaceAllCells exportedCommercialSpaceAvailability;
		IHouseholdNumberAllCells exportedVacantAccomodation;
		private IFloorSpaceAllCells exportedVacantOfficeSpace;
		private IBooleanAllCells exportedAvailabilityOfMainRoads;
		private IBooleanAllCells exportedAvailabilityOfPublicTransport;
		private IBooleanAllCells exportedSchoolSupply;
		// from infrastructure (this.getData)
		private IBooleanAllCells importedAvailabilityOfPublicTransport;
		// from infrastructure (this.getData)
		private IBooleanAllCells importedAvailabilityOfMainRoads;
		// from infrastructure (this.getData)
		private IBooleanAllCells importedSchoolSupply;
		// from infrastructure (this.getData)
		private Map<GridPoint, CLCover> importedLandUseChangesByInfrastructure;
		// from resident (this.getData)
		private Map<GridPoint, Map<Householdtypes, IHouseholdNumber>> importedTotalHouseholdNumber;
		// from business (this.getData)
		// private IFloorSpaceAllCells importedFloorSpaceOccupation;
		// from developer (this.getData)
		// private Map<GridPoint, CLCover> importedLandUseChangesByDeveloper;
		// from business (this.getData)
		// private Map<GridPoint, CLCover> importedLandUseChangesByBusiness;
		private MiscConf miscConfig;
		private int[] areaDims;
		// private IBusinessModel businessserver;
		// private IDeveloperModel developerserver;

		private IInfrastructureModel infrastructureserver;
		private IResidentModel residentserver;
		private static final String LAND_USE_EXPORTED = "Exported Land Use";
		private static final String VACANT_ACCOMODATION_TEMP = "Potential Vacant Accomodation";
		private static final String COMMERCIAL_SPACE_AVAILABLE_TEMP = "Potential Commercial Space Available";
		private static final String VACANT_OFFICE_SPACE_TEMP = "Potential Vacant Office Space";
		private static final String SCHOOL_SUPPLY_EXPORTED = "Exported School Supply";
		private static final String VACANT_ACCOMODATION_EXPORTED = "Exported Vacant Accomodation";
		private static final String COMMERCIAL_SPACE_AVAILABLE_EXPORTED = "Exported Commercial Space Available";
		private static final String VACANT_OFFICE_SPACE_EXPORTED = "Exported Vacant Office Space";
		private static final String AVAILABILITY_OF_PUBLIC_TRANSPORT_EXPORTED = "Exported Availability Of Public Transport";
		private static final String AVAILABILITY_OF_MAIN_ROADS_EXPORTED = "Exported Availability Of Main Roads";
	}

	public Env getEnv() {
		return env;
	}

	public LandUseController() {
		log.trace(null);
		// RepastUtils.getMasterContext().addSubContext(cells);
		// setLandUseMap(createMap(10, 10, null));
		// makeSimpleCity();
		// recreateLandUseDataStructures();

		env.config = ConfManager.LANDUSECONF.getConfig(LanduseConf.class);
		env.miscConfig = ConfManager.MISCCONF.getConfig();
		env.areaDims = env.miscConfig.areaParameters.getDims();

		try {
			Raster PARAM_LAND_USE_RASTER = env.config.landuseParameters
					.getRaster();

			env.PARAM_INITIAL_CLC_MAP = raster2clcmap(PARAM_LAND_USE_RASTER);

			env.exportedLandUse = requestCLCoverAllCells(PARAM_LAND_USE_RASTER,
					"landuse");

			double w = env.exportedLandUse.getDimensions().getWidth();
			double h = env.exportedLandUse.getDimensions().getHeight();

			assertTrue(w == env.areaDims[0] && h == env.areaDims[1]);

			log.trace(String.format("testing value at %s, %s (upper right corner patch): %s", w, h, CLCover
					.valueOf(env.exportedLandUse.get(w - 1, h - 1))));

		} catch (Throwable t) {
			handleException(getClass(), t);
		}
	}

	@ScheduledMethod(//
	start = 1, //
	priority = MAINCTRL_INIT_SCHED_P //
	)
	/** {@inheritDoc} */
	public void init() {
		log.trace(null);

		// env.businessserver = new BusinessModelServer2(
		// getModelController(IBusinessModel.class));
		// env.developerserver = new DeveloperModelServer2(
		// getModelController(IDeveloperModel.class));

		env.infrastructureserver = new InfrastructureModelServer2(
				getModelController(IInfrastructureModel.class));
		env.residentserver = new ResidentModelServer2(
				getModelController(IResidentModel.class));

		// land use needs to export something first... take it from a random
		// generator, the other agents haven't produced anything yet in init
		// phase
		env.exportedSchoolSupply = requestBooleanAllCells(
				Env.SCHOOL_SUPPLY_EXPORTED, true);

		env.exportedAvailabilityOfMainRoads = requestBooleanAllCells(
				Env.AVAILABILITY_OF_MAIN_ROADS_EXPORTED, true);

		env.exportedAvailabilityOfPublicTransport = requestBooleanAllCells(
				Env.AVAILABILITY_OF_PUBLIC_TRANSPORT_EXPORTED, true);

		env.exportedVacantOfficeSpace = requestFloorSpaceAllCells(
				Env.VACANT_OFFICE_SPACE_TEMP, 1);

		env.exportedCommercialSpaceAvailability = requestFloorSpaceAllCells(
				Env.COMMERCIAL_SPACE_AVAILABLE_TEMP, 1);

		env.exportedVacantAccomodation = requestHouseholdNumberAllCells(
				Env.VACANT_ACCOMODATION_TEMP, 1);
		return;
	}

	@ScheduledMethod(//
	start = 2, //
	interval = 1.0, //
	priority = MAINCTRL_GETDATA_SCHED_P //
	)
	/** {@inheritDoc} */
	public void getData() {
		log.trace(null);

		env.importedLandUseChangesByInfrastructure = env.infrastructureserver
				.getLandUseChanges();
		env.importedSchoolSupply = env.infrastructureserver.getSchoolSupply();
		env.importedAvailabilityOfPublicTransport = env.infrastructureserver
				.getAvailabilityOfPublicTransport();
		env.importedAvailabilityOfMainRoads = env.infrastructureserver
				.getAvailabilityOfMainRoads();

		env.importedTotalHouseholdNumber = env.residentserver
				.getTotalHouseholdNumber();
		return;
	}

	/**
	 * Aim of this phase:
	 *
	 * Sub A. Copy values that do not need to be changed or checked.
	 *
	 * Sub B. Update land use and other mappings.
	 *
	 * @FIXME {p:Paper} Scheduling -> land use controller has to be ready before
	 *        cost controller starts, because the latter needs updated
	 *        information. Do not allow them to compute in parallel. ...
	 *        Priorities
	 */
	@ScheduledMethod(//
	start = 2, //
	interval = 1.0, //
	priority = MAINCTRL_COMPUTE_SCHED_P //
	)
	/** {@inheritDoc} */
	public void compute() {
		log.trace(null);

		// unchanged by land use
		env.exportedSchoolSupply = env.importedSchoolSupply;
		// unchanged by land use
		env.exportedAvailabilityOfMainRoads = env.importedAvailabilityOfMainRoads;
		// unchanged by land use
		env.exportedAvailabilityOfPublicTransport = env.importedAvailabilityOfPublicTransport;
		env.exportedLandUse = updateLandUse();
		// env.exportedVacantOfficeSpace =
		// computeFloorSpace(Env.VACANT_OFFICE_SPACE_EXPORTED);
		// env.exportedCommercialSpaceAvailability =
		// computeFloorSpace(Env.COMMERCIAL_SPACE_AVAILABLE_EXPORTED);
		env.exportedVacantAccomodation = computeAccomodations(Env.VACANT_ACCOMODATION_EXPORTED);
		// updateLandUse2();

		return;
	}

	// // used in init
	// private void setLandUseMap(Raster map) {
	// Grid<Cell> cells = createMap(map.getCols(), map.getRows(),
	// CLCover.NO_DATA);
	// for (int i = 0; i < map.getCols(); i++)
	// for (int j = 0; j < map.getRows(); j++)
	// cells.getObjectAt(i, j).setLandUse(
	// CLCover.valueOf(map.getValue(i, j)));
	// }

	// // used in init
	// private void setLandUseMap(Grid<Cell> map) {
	// this.landUseMap = map;
	// recreateLandUseDataStructures();
	// }

	// // used in setLandUseMap -> Raster -> void
	// private Grid<Cell> createMap(int width, int height, CLCover fill) {
	// // cells.clear();
	// GridFactory factory = GridFactoryFinder.createGridFactory(null);
	// GridBuilderParameters<Cell> gbp = new GridBuilderParameters<Cell>(
	// new StrictBorders(), new SimpleGridAdder<Cell>(), false, width,
	// height);
	// Grid<Cell> map = (Grid<Cell>) factory.createGrid("LandUse", cells, gbp);
	// RepastUtils.getMasterContext().addProjection((Grid) map);
	// for (int i = 0; i < width; i++)
	// for (int j = 0; j < height; j++) {
	// Cell cell = new Cell();
	// cells.add(cell);
	// cell.setLandUse((fill == null ? CLCover.randomValue() : fill));
	// map.moveTo(cell, i, j);
	// }
	//
	// return map;
	// }

	// // used in init
	// private void recreateLandUseDataStructures() {
	// int x = landUseMap.getDimensions().getWidth();
	// int y = landUseMap.getDimensions().getHeight();
	// System.err.println("Going through map with dimensions " + x + "," + y);
	// GridValueLayer layer = new GridValueLayer("LandUse", true, x, y);
	// env.exportedLandUse = new CLCoverAllCells(layer);
	// for (int i = 0; i < x; i++)
	// for (int j = 0; j < y; j++)
	// env.exportedLandUse.set(landUseMap.getObjectAt(i, j)
	// .getLandUse(), i, j);
	// }

	// // used in init
	// private void makeSimpleCity() {
	// int width = landUseMap.getDimensions().getWidth();
	// int height = landUseMap.getDimensions().getWidth();
	// int cx = width / 2;
	// int cy = height / 2;
	// double r1 = 0.4;
	// for (int i = 0; i < width; i++)
	// for (int j = 0; j < height; j++) {
	// CLCover c = CLCover.LAND_PRINCIPALLY_OCCUPIED_BY_AGRICULTURE;
	// double dist = ((double) abs(i - cx) + abs(j - cy)) / (cx + cy);
	// if (dist < r1) {
	// if (RandomHelper.nextDoubleFromTo(0, r1) > dist)
	// c = CLCover.CONTINUOUS_URBAN_FABRIC;
	// else
	// c = CLCover.DISCONTINUOUS_URBAN_FABRIC;
	// } else {
	// if (RandomHelper.nextDoubleFromTo(r1, 1) > dist)
	// c = CLCover.DISCONTINUOUS_URBAN_FABRIC;
	// }
	// landUseMap.getObjectAt(i, j).setLandUse(c);
	// }
	// }

	// // depends on business exporting floor space occupation
	// private IFloorSpaceAllCells computeFloorSpace(String layerName) {
	//
	// IFloorSpaceAllCells availableFloorspace = new FloorSpaceAllCells(0,
	// null);
	// int occupiedFloorSpace;
	// ISpace availableFloorSpacePerCell;
	// int maxFloorSpacePerCell;
	// int result = 0;
	//
	// for (int x = 0; x < GetSimulationDimensions().getWidth(); x++) {
	// for (int y = 0; y < GetSimulationDimensions().getHeight(); y++) {
	// GridPoint location = new GridPoint(x, y);
	// occupiedFloorSpace = (int) env.importedFloorSpaceOccupation.get(x,
	// y);
	//
	// /*
	// * The maximum commercial space value of now comes from the
	// * templates A template consists of capacity values mapped to a
	// * CLC value
	// */
	//
	// maxFloorSpacePerCell = env.config.landuseParameters
	// .getCommercialCapacity(CLCover
	// .valueOf(env.exportedLandUse.get(x, y)));
	//
	// result = maxFloorSpacePerCell - occupiedFloorSpace;
	//
	// if (result < 0) {
	// availableFloorSpacePerCell = new Space(0);
	// log.debug(availableFloorspace.getName() + " on cell "
	// + location + " is " + availableFloorSpacePerCell);
	//
	// } else {
	// availableFloorSpacePerCell = new Space(result);
	// log.debug(availableFloorspace.getName() + " on cell "
	// + location + " is " + availableFloorSpacePerCell);
	// }
	// availableFloorspace.set(availableFloorSpacePerCell, x, y);
	// }
	// }
	// return availableFloorspace;
	// }

	private IHouseholdNumberAllCells computeAccomodations(String layerName) {

		IHouseholdNumberAllCells availableHousing = requestHouseholdNumberAllCells(
				layerName, 0);
		int totalNumberOfHouseholds;
		IHouseholdNumber availableHouseholdNumberPerCell;
		int maxHouseholdNumberPerCell;
		int result = 0;

		for (int x = 0; x < env.areaDims[0]; x++) {
			for (int y = 0; y < env.areaDims[1]; y++) {
				totalNumberOfHouseholds = 0;
				GridPoint location = new GridPoint(x, y);

				/*
				 * compute the total hdnum on the current cell
				 */
				for (Entry<Householdtypes, IHouseholdNumber> entry : env.importedTotalHouseholdNumber
						.get(location).entrySet()) {
					totalNumberOfHouseholds += entry.getValue().value();
				}

				/* query the max. hdnum for CLC on the current cell */
				maxHouseholdNumberPerCell = env.config.landuseParameters
						.getResidentialCapacity(CLCover
								.valueOf(env.exportedLandUse.get(x, y)));

				result = maxHouseholdNumberPerCell - totalNumberOfHouseholds;

				if (result < 0) {
					availableHouseholdNumberPerCell = new HouseholdNumber(0);
//					log.trace("available household number on cell " + location
//							+ " is " + availableHouseholdNumberPerCell);
				} else {
					availableHouseholdNumberPerCell = new HouseholdNumber(
							result);
//					log.trace("available household number on cell " + location
//							+ " is " + availableHouseholdNumberPerCell);
				}
				availableHousing.set(availableHouseholdNumberPerCell, x, y);
			}
		}

		return availableHousing;
	}

	// private void updateLandUse2() {
	// // Copy in the changes from the developers (we're only listening to them
	// // at the moment)
	// for (Entry<GridPoint, CLCover> cell : importedLandUseChangesByDeveloper
	// .entrySet()) {
	// System.err.println("LU Controller changing " + cell.getKey()
	// + " to " + cell.getValue());
	// int[] coords = cell.getKey().toIntArray(null);
	// landUseMap.getObjectAt(coords).setLandUse(cell.getValue());
	// env.exportedLandUse.set(cell.getValue(), coords);
	// }
	// // Update any more details for the landUseMap
	// }

	private ICLCoverAllCells updateLandUse() {

		ICLCoverAllCells newLandUse = requestCLCoverAllCells(
				Env.LAND_USE_EXPORTED, CLCover.NO_DATA);

		// Check conflicts and create mappings for Gridpoint and a) compatible
		// changes, b) conflicting changes
		// ... we could use an enumeration for agents
		Map<GridPoint, CLCover> mapOfCompatibleLUC = new HashMap<GridPoint, CLCover>();
		Map<GridPoint, HashSet<String>> mapOfChangingAgents = new HashMap<GridPoint, HashSet<String>>();

		// for (GridPoint cell : importedLandUseChangesByBusiness.keySet()) {
		// mapOfChangingAgents.get(cell).add("business");
		// }
		// for (GridPoint cell : importedLandUseChangesByDeveloper.keySet()) {
		// mapOfChangingAgents.get(cell).add("developer");
		// }

		
		
		for (GridPoint cell : env.importedLandUseChangesByInfrastructure
				.keySet()) {
			if (!mapOfChangingAgents.containsKey(cell)) {  
				mapOfChangingAgents.put(cell, new HashSet<String>());
				mapOfChangingAgents.get(cell).add("infrastructure");
			} else {
				mapOfChangingAgents.get(cell).add("infrastructure");
			}
		}

		for (GridPoint cell : mapOfChangingAgents.keySet()) {

			if (mapOfChangingAgents.get(cell).size() > 1) {
				mapOfCompatibleLUC.put(cell, resolveConflicts(cell,
						mapOfChangingAgents.get(cell)));
			} else if (mapOfChangingAgents.get(cell).size() == 1) {
				CLCover _ = null;
				try {
					_ = insertCompatibleLandUse(cell,
							mapOfChangingAgents.get(cell));
					_.value();
					mapOfCompatibleLUC.put(cell, _);
				} catch (Throwable e) {
					handleException(LandUseController.class, e);
					System.exit(1564);
				}
			}
		}

		newLandUse = summariseLandUse(newLandUse, mapOfCompatibleLUC);

		return newLandUse;
	}

	/**
	 * Is called if more than 1 agent wants to change a land use on a certain
	 * cell. Conflict solving: priority of agents: 1. business. 2. developer. 3.
	 * infrastructure.
	 *
	 * @param location
	 * @param importedLUCByBusiness
	 * @param importedLUCByDeveloper
	 * @param importedLUCByInfrastructure
	 * @param changingAgents
	 * @return CLCover -> one land use per cell
	 */
	private CLCover resolveConflicts(GridPoint location,
			HashSet<String> changingAgents) {

		log.debug("confict on cell " + location.toString() + " with agents "
				+ changingAgents.toString());

		CLCover landUse = null;

		/**
		 * if (changingAgents.contains("business")) { landUse =
		 * importedLandUseChangesByBusiness.get(location); } else if
		 * (changingAgents.contains("developer")) { landUse =
		 * importedLandUseChangesByDeveloper.get(location); } else
		 */
		if (changingAgents.contains("infrastructure")) {
			landUse = env.importedLandUseChangesByInfrastructure.get(location);
		} else {
			log.debug("unknown agent which changes land use");
		}
		return landUse;

	}

	private CLCover insertCompatibleLandUse(GridPoint location,
			HashSet<String> changingAgents) {

		CLCover landUse = null;

		/**
		 * if (changingAgents.toString() == "business") { landUse =
		 * importedLandUseChangesByBusiness.get(location); } else if
		 * (changingAgents.toString() == "developer") { landUse =
		 * importedLandUseChangesByDeveloper.get(location); } else
		 */
		for (String string : changingAgents) {
			if (string == "infrastructure") {
				landUse = env.importedLandUseChangesByInfrastructure.get(location);
			} else {
				log.debug("unknown agent which changes land use, agent = " + changingAgents.toString());
			}
		}
		return landUse;
	}

	private ICLCoverAllCells summariseLandUse(ICLCoverAllCells newLandUse,
			Map<GridPoint, CLCover> compatibleLUC) {

		// FIXME {p:Paper} ??? Parameter overridden immediately, maybe we don't
		// need it in general?
		newLandUse = env.exportedLandUse; // new = land use from last time step

		for (GridPoint location : compatibleLUC.keySet()) {
			// FIXME {p:Paper} what happens if there is already a value on that
			// cell? No possibility to remove that beforehand?
			// FIXME Fail fast test if luc is null
			newLandUse.set(compatibleLUC.get(location), location.getX(),
					location.getY());
		}
		return newLandUse;
	}

	/**
	 * TODO This is requested by our interface table... {@inheritDoc}
	 */
	public CLCover getScalarLandUse() {
		return CLCover.NO_DATA;
	}

	/** {@inheritDoc} */
	public IFloorSpaceAllCells getVacantOfficeSpace() {
		// currently returned as imported in init phase
		return env.exportedVacantOfficeSpace;
	}

	/** {@inheritDoc} */
	// [2] 15.09.10 - 13:49:12 Uhr
	// FIXME {p:Paper} Should resident query this land use ctrl rather instead
	// of infrastructure
	public IBooleanAllCells getAvailabilityOfMainRoads() {
		// currently returned as imported in init phase
		return env.exportedAvailabilityOfMainRoads;
	}

	/** {@inheritDoc} */
	// FIXME {p:Paper} Should resident query this land use ctrl rather instead
	// of infrastructure
	public IBooleanAllCells getAvailabilityOfPublicTransport() {
		// currently returned as imported in init phase
		return env.exportedAvailabilityOfPublicTransport;
	}

	/** Currently not computed. {@inheritDoc} */
	public IFloorSpaceAllCells getCommercialSpaceAvailability() {
		// currently returned as imported in init phase
		return env.exportedCommercialSpaceAvailability;
	}

	/**
	 * @FIXME {p:Paper} Make a canonic path from the existing method stubs to
	 *        compute that, for a start only with infrastructure land use
	 *        request. {@inheritDoc}
	 */
	public ICLCoverAllCells getLandUse() {
		return env.exportedLandUse;
	}

	// @Parameter(displayName = "school supply", usageName = "schoolSupply")
	/** {@inheritDoc} */
	// FIXME {p:Paper} Should resident query this land use ctrl rather instead
	// of infrastructure
	public IBooleanAllCells getSchoolSupply() {
		// currently returned as imported in init phase
		return env.exportedSchoolSupply;
	}

	/** {@inheritDoc} */
	public IHouseholdNumberAllCells getVacantAccomodation() {
		return env.exportedVacantAccomodation;
	}
}

//class Cell
//{
//	CLCover landUse;
//	int commercialCapacity;
//	int residentialCapacity;
//
//	int commercialUsage;
//	int residentialUsage;
//
//
//	public CLCover getLandUse()
//	{
//		return landUse;
//	}
//	public void setLandUse( CLCover landUse )
//	{
//		this.landUse = landUse;
//	}
//
//	public int getCommercialUsage()
//	{
//		return commercialUsage;
//	}
//	public void setCommercialUsage( int commercialUsage )
//	{
//		this.commercialUsage = commercialUsage;
//	}
//	public int getResidentialUsage()
//	{
//		return residentialUsage;
//	}
//	public void setResidentialUsage( int residentialUsage )
//	{
//		this.residentialUsage = residentialUsage;
//	}
//	public int getCommercialCapacity()
//	{
//		return commercialCapacity;
//	}
//	public int getResidentialCapacity()
//	{
//		return residentialCapacity;
//	}
//
//	public String getLUName()
//	{
//		return landUse.toString();
//	}
//
//
//	public String toString()
//	{
//		if( landUse == null ) return "-1";
//		return (int)landUse.value() + "";
//	}
//}

