/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.models.config;

import static abmland.common.Properties.CASESTUDY;
import static abmland.common.Properties.DECISIONMODE;
import static abmland.common.Properties.SCENARIONAME;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Method;

import org.apache.log4j.Logger;
import org.simpleframework.xml.core.Persister;

import repast.simphony.engine.environment.RunEnvironment;

import com.moseph.gis.raster.Raster;
import com.moseph.gis.raster.RasterReader;

/**
 * Reads in config files based on the case study, scenario etc. specified with
 * the extended parameters and model.score.
 *
 * Asks also for a ModelConfig subclass, which defines the type of the model to
 * be created, and is used to generate the filename.
 *
 * @author dave
 *
 */
public class ConfigReader extends Persister {

	private static String DS = System.getProperty("file.separator");

	static Logger log = Logger.getLogger(ConfigReader.class);

	ClassLoader classLoader;
	String baseDir = null;

	public ConfigReader() {
		classLoader = getClass().getClassLoader();
	}

	public ConfigReader(ClassLoader classLoader) {
		this.classLoader = classLoader;
	}

	// Sorry about the classloader stuff, it's uglier than it should be!
	// There's a bad interaction with the Repast bootloader system, which means
	// that the model's classes are not available in the classloader for this
	// thread, but they are there for this class.
	/**
	 * {@inheritDoc}
	 */
	public <T> T read(Class<? extends T> type, File source) throws Exception {
		ClassLoader defaultLoader = Thread.currentThread()
				.getContextClassLoader();
		try {
//			System.err.println("Munging classloader!");
			log.trace("Munging classloader!");
			Thread.currentThread().setContextClassLoader(classLoader);
			return super.read(type, source);
		} finally {
			Thread.currentThread().setContextClassLoader(defaultLoader);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public <T> T read(Class<? extends T> type, InputStream source)
			throws Exception {
		ClassLoader defaultLoader = Thread.currentThread()
				.getContextClassLoader();
		try {
//			System.err.println("Munging classloader!");
			log.trace("Munging classloader!");
			Thread.currentThread().setContextClassLoader(classLoader);
			return super.read(type, source);
		} finally {
			Thread.currentThread().setContextClassLoader(defaultLoader);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public <T> T read(Class<? extends T> type, String filename)
			throws Exception {
		return read(type, new File(filename));
	}

	public <T extends ModelConfig> T readConfig(String caseStudy,
			String scenario, String decisionType, Class<T> modelType)
			throws Exception {
		String filename = generateFilename(caseStudy, scenario, decisionType,
				modelType);
		File f = new File(filename);

		log.trace("(Var1) Reading config for " + modelType.getSimpleName()
				+ " from file: " + f.getAbsolutePath());

		try {
			log
					.trace("(Var2) Reading config for "
							+ modelType.getSimpleName()
							+ " from file: "
							+ new File(getClass().getClassLoader().getResource(
									filename).getFile()).getAbsolutePath());
		} catch (Exception e) {
			for (StackTraceElement st : e.getStackTrace()) {
				log.trace(st);
			}
		}
		try {
			log.trace("(Var3) Reading config for " + modelType.getSimpleName()
					+ " from input stream");
		} catch (Exception e) {
			for (StackTraceElement st : e.getStackTrace()) {
				log.trace(st);
			}
		}
		return read(modelType, getClass().getClassLoader().getResourceAsStream(
				filename));
	}

	public <T extends ModelConfig> T readConfig(Class<T> modelType)
			throws Exception {
		String scenario = RunEnvironment.getInstance().getParameters()
				.getValue(SCENARIONAME).toString();
		String caseStudy = RunEnvironment.getInstance().getParameters()
				.getValue(CASESTUDY).toString();
		String decisionType = RunEnvironment.getInstance().getParameters()
				.getValue(DECISIONMODE).toString();
		return readConfig(caseStudy, scenario, decisionType, modelType);
	}

	public static String generateFilename(String caseStudy, String scenario,
			String decisionType, Class<? extends ModelConfig> model) {
		String confName = "unknown";
		Method m;
		try {

			m = model.getMethod("getModelConfigName");
			/**
			 * FIXME Exception should be thrown when getModelConfigName returns
			 * null
			 */
			/**
			 * FIXME newInstance call seems cost-intensive, but if we used an
			 * object as enclosing method's parameter, we may get in trouble
			 * with all the other read, readConfig methods...
			 */
			confName = m.invoke(model.newInstance()).toString();
		} catch (NullPointerException npe) {
			// confName = m.invoke(null).toString();
			log.error(npe.getMessage() + '\n' + '\t' + npe.getStackTrace()[0]
					+ "...");
		} catch (IllegalArgumentException iae) {
			// confName = m.invoke(model).toString();
			log.error(iae.getMessage() + '\n' + '\t' + iae.getStackTrace()[0]
					+ "...");
		} catch (Exception e) {
			log.error(e.getStackTrace()[0]);
			e.printStackTrace();
			confName = model.getSimpleName().replaceAll("Model", "")
					.replaceAll("Config", "").toLowerCase();
		}
		return "config" + DS + caseStudy + DS + scenario + DS + decisionType
				+ "." + confName + ".xml";

	}

	public static String generateInitialMapFilename() {
		String caseStudy = RunEnvironment.getInstance().getParameters()
				.getValueAsString("case_study_name");
		return "config" + DS + caseStudy + DS + "initialMap.asc";
	}

	public Raster getInitialMap() throws IOException {
		return new RasterReader().readRaster(generateInitialMapFilename());
	}
}
