/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.models.config;

import static abmland.common.LogExceptionHandler.handleException;
import static abmland.common.Properties.CASESTUDY;
import static abmland.common.Properties.STARTYEAR;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.fail;
import static abmland.mainctrl.ParametersController.getYParameters;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.lang.reflect.Method;

import org.apache.log4j.Logger;
import org.simpleframework.xml.core.Persister;


/**
 * This is the version of Daves config reader without the class loading proxy
 * stuff.
 *
 * @author dave
 *
 */
public class ConfigReaderImpl extends Persister {

	private static String DS = System.getProperty("file.separator");

	private static String RESOURCE_COULD_NOT = "Resource could not be found: %s";

	private static String TL_RESOURCE_COULD = "%s  not found, but resource top folder is found, so double-check your config values in extended_params.xml, e. g. STARTYEAR.";

	static Logger log = Logger.getLogger(ConfigReaderImpl.class);

	ClassLoader classLoader;
	String baseDir = null;
	static String mapsConfig = "config" + DS + "maps";
	static String paramsConfig = "config" + DS + "params";
	static String testResource = "config" + DS + "params";

	public ConfigReaderImpl() {
		classLoader = getClass().getClassLoader();
	}

	public ConfigReaderImpl(ClassLoader classLoader) {
		this.classLoader = classLoader;
	}

	/**
	 * Read by input stream {@inheritDoc}
	 */
	public <T> T read(Class<? extends T> type, InputStream inputstream)
			throws Exception {

		return super.read(type, inputstream);
	}

	public <T extends ModelConfig> T readConfig(String configDirPath,
			Class<T> modelType) /** throws Exception */
	{
		String configFilename = generateFilename(configDirPath, modelType);
		assertNotNull("file name not generated: " + configFilename,
				configFilename);

		File file = new File(configFilename);

		ClassLoader cll = Thread.currentThread().getContextClassLoader();
//		ClassLoader cll = ClassLoader.getSystemClassLoader();

		/**
		 * Possibilities (trial-error):
		 *
		 * ClassLoader.getSystemClassLoader();
		 * ClassLoader.getSystemResource(name);
		 * Thread.currentThread().getContextClassLoader();
		 * ClassLoader.getSystemClassLoader().getResource(name)
		 * ClassLoader.getSystemClassLoader().getParent().getResource(name)
		 */
		try {
			assertNotNull(String.format(RESOURCE_COULD_NOT, testResource), cll
					.getResource(testResource));
			assertNotNull(String.format(TL_RESOURCE_COULD, configFilename), cll
					.getResource(configFilename));
		} catch (Throwable t) {
//			handleException(getClass(), t);
		}

		// (1) Try to read as file...
		try {
			// [debug] throws FileNotFoundException
			new FileInputStream(file);
			log.trace("Delegating to read per file (" + configFilename
					+ " found).");
			return super.read(modelType, file);
		} catch (Throwable t) {
			// handleException(getClass(), t);
			// (2) Try to read as input stream...
			try {
				ClassLoader cl = getClass().getClassLoader();
				InputStream configFileIs = cl
						.getResourceAsStream(configFilename);
				InputStream topLevelIs = cl.getResourceAsStream(testResource);

				assertNotNull(RESOURCE_COULD_NOT + testResource, topLevelIs);
				assertNotNull(String.format(TL_RESOURCE_COULD, configFilename),
						configFileIs);
				log.trace("Delegating to read per input stream ("
						+ configFilename + " found).");
				return super.read(modelType, configFileIs);
			} catch (Throwable t2) {
				handleException(getClass(), t2);
				fail();
			}
		}

		// otherwise return nothing
		return null;
	}

	public static String generateFilename(String configFileBasedir,
			Class<? extends ModelConfig> model) {
		String confName = "unknown";
		Method m;
		try {

			m = model.getMethod("getModelConfigName");
			/**
			 * FIXME Exception should be thrown when getModelConfigName returns
			 * null
			 */
			/**
			 * FIXME newInstance call seems cost-intensive, but if we used an
			 * object as enclosing method's parameter, we may get in trouble
			 * with all the other read, readConfig methods...
			 */
			confName = m.invoke(model.newInstance()).toString();
		} catch (NullPointerException npe) {
			// confName = m.invoke(null).toString();
			log.error(npe.getMessage() + '\n' + '\t' + npe.getStackTrace()[0]
					+ "...");
		} catch (IllegalArgumentException iae) {
			// confName = m.invoke(model).toString();
			log.error(iae.getMessage() + '\n' + '\t' + iae.getStackTrace()[0]
					+ "...");
		} catch (Exception e) {
			log.error(e.getStackTrace()[0]);
			e.printStackTrace();
			confName = model.getSimpleName().replaceAll("Model", "")
					.replaceAll("Config", "").toLowerCase();
		}
		return configFileBasedir + DS + confName + ".xml";
	}
}
