/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.models.config;

import static abmland.common.LogExceptionHandler.handleException;
import static org.junit.Assert.assertNotNull;
import static org.apache.log4j.Logger.getLogger;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import repast.simphony.space.grid.GridPoint;

import abmland.data.CLCover;
import abmland.types.IPriceIndex;
import abmland.types.impl.PriceIndex;

import com.moseph.gis.raster.Raster;

public class RasterIO {

	private static final Logger log = getLogger(RasterIO.class);
	private static final String RASTERNULL = "Raster creation seems to have failed: %s";

	public static Raster buildRaster(String filename) {
		Raster raster = null;
		try {
			raster = new RasterReaderWrapper().readRaster(filename);
			assertNotNull(String.format(RASTERNULL, raster), raster);
		} catch (IOException e) {
			handleException(RasterIO.class, e);
			fail();
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return raster;
	}

	public static int[] getRasterDimensions(Raster raster) {
		try {
			return new int[] { raster.getRows(), raster.getCols() };
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return null;
	}

	public static Map<GridPoint, Float> raster2fmap(Raster raster) {
		Map<GridPoint, Float> map = null;
		try {
			map = new HashMap<GridPoint, Float>();
			for (int rows = 0; rows < raster.getRows(); rows++) {
				for (int cols = 0; cols < raster.getCols(); cols++) {
					double v = raster.getValue(rows, cols);
					try {
						boolean isNoData = Double.isNaN(v);
						assertFalse(
								String
										.format(
												"No grid-point boolean pair stored: %s (%s,%s)",
												v, rows, cols), isNoData);
						map.put(new GridPoint(rows, cols), (float) v);
					} catch (Throwable e) {// this is really only debug info. if
						// no gridpoint is
						// stored, anyway, as long as we had -9999 in the ascii
						// log.debug(e.getMessage());
					}
				}
			}
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}

	public static Map<GridPoint, CLCover> raster2clcmap(Raster raster) {
		Map<GridPoint, CLCover> map = new HashMap<GridPoint, CLCover>();
		try {
			for (int rows = 0; rows < raster.getRows(); rows++) {
				for (int cols = 0; cols < raster.getCols(); cols++) {
					double v = raster.getValue(rows, cols);
					try {
						boolean isNoData = Double.isNaN(v);
						assertFalse(
								String
										.format(
												"No grid-point boolean pair stored: %s (%s,%s)",
												v, rows, cols), isNoData);
						map.put(new GridPoint(rows, cols), CLCover.valueOf(v));
					} catch (Throwable e) {// this is really only debug info. if
						// no gridpoint is
						// stored, anyway, as long as we had -9999 in the ascii
						// log.debug(e.getMessage());
					}
				}
			}
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}

	/**
	 * FIXME {p:Paper Raster can contain NODATA points, they are stored by the current
	 * implementation Double.NaN which isn't cross-datatype (we
	 * had Integer too, which hasn't NaN constant, now we use Double). If such a value is detected
	 * should it be stored here and how (as Double.parseDouble(Raster.getNDATA)?) within the resulting map? This is a paradigmatic question having consequences on the constraints handling our java maps, so in short again: Should those java maps store NODATA points too?}
	 *
	 * @param raster
	 * @return
	 */
	// later on we could use the boundedfloat/priceindex
	public static Map<GridPoint, IPriceIndex> raster2priceindexmap(Raster raster) {
		Map<GridPoint, IPriceIndex> map = new HashMap<GridPoint, IPriceIndex>();
		try {
			for (int rows = 0; rows < raster.getRows(); rows++) {
				for (int cols = 0; cols < raster.getCols(); cols++) {
					double v = raster.getValue(rows, cols);
					try {
					boolean isNoData = Double.isNaN(v);
						assertFalse(
								String
										.format(
												"No grid-point boolean pair stored: %s (%s,%s)",
												v, rows, cols), isNoData);
						map.put(new GridPoint(rows, cols), new PriceIndex(v));
					} catch (Throwable e) {// this is really only debug info. if
						// no gridpoint is
						// stored, anyway, as long as we had -9999 in the ascii
						// log.debug(e.getMessage());
					}
				}
			}
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}

	public static Map<GridPoint, Number> raster2nmap(Raster raster) {
		Map<GridPoint, Number> map = new HashMap<GridPoint, Number>();
		try {
			for (int rows = 0; rows < raster.getRows(); rows++) {
				for (int cols = 0; cols < raster.getCols(); cols++) {
					double v = raster.getValue(rows, cols);
					try {
						boolean isNoData = Double.isNaN(v);
						assertFalse(
								String
										.format(
												"No grid-point boolean pair stored: %s (%s,%s)",
												v, rows, cols), isNoData);
						map.put(new GridPoint(rows, cols), v);
					} catch (Throwable e) {// this is really only debug info. if
						// no gridpoint is
						// stored, anyway, as long as we had -9999 in the ascii
						// log.debug(e.getMessage());
					}
				}
			}
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}

	public static Map<GridPoint, Integer> raster2imap(Raster raster) {
		Map<GridPoint, Integer> map = new HashMap<GridPoint, Integer>();
		try {
			for (int rows = 0; rows < raster.getRows(); rows++) {
				for (int cols = 0; cols < raster.getCols(); cols++) {
					double v = raster.getValue(rows, cols);
					try {
						boolean isNoData = Double.isNaN(v);
						assertFalse(
								String
										.format(
												"No grid-point boolean pair stored: %s (%s,%s)",
												v, rows, cols), isNoData);
						map.put(new GridPoint(rows, cols), (int) v);
					} catch (Throwable e) {// this is really only debug info. if
						// no gridpoint is
						// stored, anyway, as long as we had -9999 in the ascii
						// log.debug(e.getMessage());
					}
				}
			}
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}

	/**
	 * @FIXME {p:Paper Sohuld that be assumed for all maps? Meaning if NODATA
	 *        given don't store map entries, that would have consequences for
	 *        methods that assume the maps having full case study area coverage,
	 *        otherwise it could improve efficiency} Raster can contain NODATA
	 *        points, they are stored by the current implementation as Float.NaN
	 *        or Double.NaN which isn't cross-datatype (we have Integer too,
	 *        which hasn't NaN constant). If such a value is detected it is not
	 *        stored as within the resulting map as there can be no Boolean
	 *        equivalent for NODATA as Boolean isn't ternary
	 *
	 * @param raster
	 * @return
	 */
	public static Map<GridPoint, Boolean> raster2boolmap(Raster raster) {
		Map<GridPoint, Boolean> map = new HashMap<GridPoint, Boolean>();
		for (int rows = 0; rows < raster.getRows(); rows++) {
			for (int cols = 0; cols < raster.getCols(); cols++) {
				double v = raster.getValue(rows, cols);
				try {
					boolean isDecision = v == 0.0 || v == 1.0;
					assertTrue(String.format(
							"No grid-point boolean pair stored: %s (%s,%s)", v,
							rows, cols), isDecision);
					map.put(new GridPoint(rows, cols), (v == 1) ? true : false);
				} catch (Throwable e) {
					// this is really only debug info. if no gridpoint is
					// stored, anyway, as long as we had -9999 in the ascii
					// log.debug(e.getMessage());
				}
				// catch wrong value in ascii errors
				try {
					boolean isValid = v == 0.0 || v == 1.0 || Double.isNaN(v);
					assertTrue(
							String
									.format(
											"Not a valid boolean equivalent (need 0 or 1) or NODATA at least: %s (%s,%s)",
											v, rows, cols), isValid);
				} catch (Throwable e) {
					log.error(e.getMessage());
				}
			}
		}
		return map;
	}

	public static Map<GridPoint, Float> raster2fmap(String filename) {
		Map<GridPoint, Float> map = null;
		try {
			Raster raster = new RasterReaderWrapper().readRaster(filename);
			assertNotNull(String.format(RASTERNULL, raster), raster);
			map = raster2fmap(raster);
		} catch (IOException e) {
			handleException(RasterIO.class, e);
			fail();
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}

	public static Map<GridPoint, Integer> raster2imap(String filename) {
		Map<GridPoint, Integer> map = null;
		try {
			Raster raster = new RasterReaderWrapper().readRaster(filename);
			assertNotNull(String.format(RASTERNULL, raster), raster);
			map = raster2imap(raster);
		} catch (IOException e) {
			handleException(RasterIO.class, e);
			fail();
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}

	public static Map<GridPoint, Boolean> raster2boolmap(String filename) {
		Map<GridPoint, Boolean> map = null;
		try {
			// FIXME readRaster should be called indirectly after resolving to
			// correct parameter set and so parameter file locations
			Raster raster = new RasterReaderWrapper().readRaster(filename);
			assertNotNull(String.format(RASTERNULL, raster), raster);
			map = raster2boolmap(raster);
		} catch (IOException e) {
			handleException(RasterIO.class, e);
			fail();
		} catch (Throwable t) {
			handleException(RasterIO.class, t);
			fail();
		}
		return map;
	}
}
