/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.types.impl;

import static abmland.data.DatatypeBoundaries.valueOf;
import static org.junit.Assert.assertTrue;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import repast.simphony.space.Dimensions;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.space.grid.GridPointTranslator;
import repast.simphony.valueLayer.GridValueLayer;
import repast.simphony.valueLayer.ValueLayer;
import abmland.common.DefaultLayer;
import abmland.common.LayerEvent;
import abmland.common.LayerListener;
import abmland.data.CLCover;
import abmland.types.ICompositeAllCells;

/**
 * See `repast.simphony.space.grid.AbstractGrid`
 *
 * @tag task.todo : Should `AbstractAllCells&lt;T&gt;` be a mirror of
 *      GridValueLayer?
 *
 * @tag usage.general : Extends `repast.simphony.valueLayer.ValueLayer` (to be
 *      usable with `addValueLayer()`) and has all methods a collaborator
 *      needed.
 *
 * @author dkahlenb
 *
 * @param <T>
 */
public abstract class AbstractAllCells<T/** extends Number */
> extends DefaultLayer<T> implements ValueLayer {

	GridValueLayer layer;

	/**
	 * @tag usage.general -priority="4" : Ensure that the layer given as
	 *      argument has the correct dimensions
	 * @param layer
	 */
	public AbstractAllCells(GridValueLayer layer) {
		this.layer = layer;
	}

	/** {@inheritDoc} */
	public double get(double... coordinates) {

		return layer.get(coordinates);
	}

	public abstract void set(T value, int... coordinate);

	public void set(double value, int... coordinate) {
		layer.set(value, coordinate);
	}

	public void layerEventOccurred(LayerEvent<T> evt) {
		Collection<LayerListener<T>> listeners = getLayerListeners();
		for (LayerListener<T> layerListener : listeners) {
			layerListener.layerEventOccurred(evt);
		}
	}

	/** {@inheritDoc} */
	public Dimensions getDimensions() {
		return layer.getDimensions();
	}

	/** {@inheritDoc} */
	public String getName() {
		return layer.getName();
	}

	protected int[] getTransformedLocation(int... location) {
		int[] loc = new int[location.length];
		layer.getGridPointTranslator().transform(loc, location);
		return loc;
	}

	public GridPointTranslator getGridPointTranslator() {
		return layer.getGridPointTranslator();
	}

	public void setGridPointTranslator(GridPointTranslator rule) {
		layer.setGridPointTranslator(rule);
	}

	/**
	 * Set all values from the matrix underlying in the `GridValueLayer` to NaN
	 */
	public void clear() {
		for (int fst = 0; fst < layer.getDimensions().toIntArray(null)[0]; fst++) {
			for (int snd = 0; snd < layer.getDimensions().toIntArray(null)[1]; snd++) {
				layer.set(Double.NaN, fst, snd);
			}
		}
	}

	public static Map<GridPoint, CLCover> allCellsToCLCoverMap(
			ICompositeAllCells<?> allCells) {
		Map<GridPoint, CLCover> map = new HashMap<GridPoint, CLCover>();

		Dimensions dimensions = allCells.getDimensions();

		switch (dimensions.size()) {
		case 2:
			for (int w = 0; w < dimensions.getWidth(); w++) {
				for (int h = 0; h < dimensions.getHeight(); h++) {
					GridPoint gp = new GridPoint(w, h);
					CLCover vl = CLCover.valueOf(allCells.get(w, h));
					map.put(gp, vl);
					assertTrue("Stored " + vl + " into map, but got "
							+ map.get(gp), vl == map.get(gp));
				}
			}
			break;
		case 3:
			for (int w = 0; w < dimensions.getWidth(); w++) {
				for (int h = 0; h < dimensions.getHeight(); h++) {
					for (int d = 0; d < dimensions.getDepth(); d++) {
						GridPoint gp = new GridPoint(w, h, d);
						CLCover vl = CLCover.valueOf(allCells.get(w, h, d));
						map.put(gp, vl);
						assertTrue("Stored " + vl + " into map, but got "
								+ map.get(gp), vl == map.get(gp));
					}
				}
			}
			break;
		default:
			break;
		}
		return map;
	}

	public static Map<GridPoint, Number> allCellsToMap(
			ICompositeAllCells<?> allCells) {
		Map<GridPoint, Number> map = new HashMap<GridPoint, Number>();

		Dimensions dimensions = allCells.getDimensions();

		switch (dimensions.size()) {
		case 2:
			for (int w = 0; w < dimensions.getWidth(); w++) {
				for (int h = 0; h < dimensions.getHeight(); h++) {
					GridPoint gp = new GridPoint(w, h);
					Number vl = allCells.get(w, h);
					map.put(gp, vl);
					assertTrue("Stored " + vl + " into map, but got "
							+ map.get(gp), vl == map.get(gp));
				}
			}
			break;
		case 3:
			for (int w = 0; w < dimensions.getWidth(); w++) {
				for (int h = 0; h < dimensions.getHeight(); h++) {
					for (int d = 0; d < dimensions.getDepth(); d++) {
						GridPoint gp = new GridPoint(w, h, d);
						Number vl = allCells.get(w, h, d);
						map.put(gp, vl);
						assertTrue("Stored " + vl + " into map, but got "
								+ map.get(gp), vl == map.get(gp));
					}
				}
			}
			break;
		default:
			break;
		}
		return map;
	}

	public static Map<GridPoint, Boolean> allCellsToBooleanMap(
			ICompositeAllCells<?> allCells) {
		Map<GridPoint, Boolean> map = new HashMap<GridPoint, Boolean>();

		Dimensions dimensions = allCells.getDimensions();

		switch (dimensions.size()) {
		case 2:
			for (int w = 0; w < dimensions.getWidth(); w++) {
				for (int h = 0; h < dimensions.getHeight(); h++) {
					GridPoint gp = new GridPoint(w, h);
					Number vl = allCells.get(w, h);
					map.put(gp, vl == valueOf(true) ? true
							: vl == valueOf(false) ? false : false);
					// assertTrue("Stored " + true + "|" + false
					// + " into map, but got " + map.get(gp), true == map
					// .get(gp)
					// || false == map.get(gp));
				}
			}
			break;
		case 3:
			for (int w = 0; w < dimensions.getWidth(); w++) {
				for (int h = 0; h < dimensions.getHeight(); h++) {
					for (int d = 0; d < dimensions.getDepth(); d++) {
						GridPoint gp = new GridPoint(w, h, d);
						Number vl = allCells.get(w, h, d);
						map.put(gp, vl == valueOf(true) ? true
								: vl == valueOf(false) ? false : false);
					}
				}
			}
			break;
		default:
			break;
		}
		return map;
	}
}
