/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.types.impl;

import static abmland.common.LayerEvent.OBJECT_ADDED;
import static abmland.common.LayerEvent.OBJECT_INVALID;
import static abmland.common.StringUtils.charr2String;
import static abmland.data.DatatypeBoundaries.valueOf;
import static org.junit.Assert.assertTrue;

import java.util.Collection;

import repast.simphony.valueLayer.GridValueLayer;
import abmland.common.LayerEvent;
import abmland.common.LayerListener;
import abmland.mainctrl.config.ConfManager;
import abmland.mainctrl.config.MiscConf;
import abmland.types.IBooleanAllCells;

public class BooleanAllCells extends AbstractAllCells<Boolean> implements
		IBooleanAllCells {

	/**
	 * @tag usage.protocol : Assumes that everything
	 *      (src/main/resources/config/params/.../MiscConf.xml) is initialized
	 *      correctly to retrieve the configuration driven default dimensions.
	 */
	private static int[] dimensions = ConfManager.MISCCONF
			.getConfig(MiscConf.class).areaParameters.getDims();

	private static final String NAME = "BooleanAllCellsGridVL";
	private final int TRUE = (Integer) valueOf(true);
	private final int FALSE = (Integer) valueOf(false);

	private void handleBoundaries(Double value) throws AssertionError {
		try {
			assertTrue(value + " not mapped to TRUE or FALSE.", (value == TRUE)
					|| (value == FALSE));
		} catch (AssertionError e) {
			layerEventOccurred(new LayerEvent<Boolean>(this, value,
					OBJECT_INVALID));
			e.printStackTrace();
		}
	}

	/**
	 * @tag usage.protocol : Assumes that the given layer has the correct
	 *      dimensions.
	 */
	public BooleanAllCells(GridValueLayer layer/** , double defaultValue */
	) {
		super(layer);
	}

	/**
	 *
	 * @param value
	 * @param name of underlying grid value layer - optional
	 */
	public BooleanAllCells(Boolean value, String... name) {
		// boundary checking inclusive
		super(new GridValueLayer(charr2String(NAME, name), valueOf(value)
				.doubleValue(), true, dimensions));
	}

	/**
	 * @param name of underlying grid value layer - optional
	 * @tag usage.protocol : Assumes that everything
	 *      (src/main/resources/config/params/.../MiscConf.xml) is initialized
	 *      correctly to retrieve the configuration driven default dimensions.
	 */
	public BooleanAllCells(String... name) {
		super(new GridValueLayer(charr2String(NAME, name), valueOf(false)
				.doubleValue(), true, dimensions));
	}

	/** {@inheritDoc} */
	public void set(Boolean value, int... coordinate) {
		this.set(valueOf(value).doubleValue(), coordinate);
	}

	/**
	 * @tag usage.sideeffect : We will raise the layer event
	 *      `Type.OBJECT_INVALID` and throw an exception, if a wrong-bounded
	 *      value would have been stored.
	 */
	/** {@inheritDoc} */
	public void set(double value, int... coordinate) {
		int intvalue = new Double(value).intValue();
		try {
			assertTrue(value + " out of borders.", (intvalue == FALSE)
					|| (intvalue == TRUE));
			/**
			 * Only store if no [exception thrown until now] boundaries broken
			 */
			super.set(value, coordinate);
			layerEventOccurred(new LayerEvent<Boolean>(this, value,
					OBJECT_ADDED));

		} catch (AssertionError e) {
			layerEventOccurred(new LayerEvent<Boolean>(this, value,
					OBJECT_INVALID));
			e.printStackTrace();
		}
	}

	/** {@inheritDoc} */
	public void layerEventOccurred(LayerEvent<Boolean> evt) {
		Collection<LayerListener<Boolean>> listeners = getLayerListeners();
		for (LayerListener<Boolean> layerListener : listeners) {
			layerListener.layerEventOccurred(evt);
		}
	}
}
