/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.types.impl;

import static org.apache.log4j.Logger.getLogger;

import org.apache.log4j.Logger;

import abmland.core.sandbox.BoundsUtil;
import abmland.types.BoundedFloat;
import fj.P1;
import fj.data.Either;

@BoundedFloat(MIN = -.5f, MAX = .5f)
public class BoundedFloatImpl extends Number implements
		Comparable<BoundedFloatImpl> {

	private static final Logger log = getLogger(BoundedFloatImpl.class);

	private float value;

	private static Class<BoundedFloat> inherited_ = BoundedFloat.class;
	private static Class<BoundedFloatImpl> this_ = BoundedFloatImpl.class;

	private static Float MIN = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MIN() : null;

	private static Float MAX = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MAX() : null;

	public static final Class<Float> TYPE = Float.TYPE;

	private BoundedFloatImpl() {
		;
	}

	public BoundedFloatImpl(float value, final float MIN, final float MAX) {
		BoundedFloatImpl.MIN = MIN;
		BoundedFloatImpl.MAX = MAX;
		this.value = guardAssignment(value, MIN, MAX);
	}

	public BoundedFloatImpl(double value, final float MIN, final float MAX) {
		BoundedFloatImpl.MIN = MIN;
		BoundedFloatImpl.MAX = MAX;
		this.value = guardAssignment((float) value, MIN, MAX);
	}

	// public BoundedFloatImpl(float value) {
	// this.value = guardAssignment(value, MIN, MAX);
	// }

	// public BoundedFloatImpl(double value) {
	// this.value = guardAssignment((float) value);
	// }

	private static float guardAssignment(float value, final float MIN,
			final float MAX) {
		float EDOM = Float.NaN;

		P1<Either<Exception, Float>> lazyE = BoundedFloatImpl.eAssign(value,
				MIN, MAX);
		float result = (lazyE._1().isLeft() ? EDOM : lazyE._1().right().value());
		if (lazyE._1().isLeft()) {
			log.error("assigned value " + value + " is out of bounds. Assigning " + EDOM, lazyE._1().left().value());
		}
		return result;
	}

	private static P1<Either<Exception, Float>> eAssign(final Float value,
			final float MIN, final float MAX) {
		return new P1<Either<Exception, Float>>() {
			/** {@inheritDoc} */
			public Either<Exception, Float> _1() {
				try {
					if (BoundsUtil.isOuterBounds(value, MIN, MAX)) {
						throw new Exception(value + " outer bounds (" + MIN + ", " + MAX + ")");
					}
					return Either.right(value);
				} catch (Exception e) {
					return Either.left(e);
				}
			}
		};
	}

	public static void main(String[] args) {
		log.info(BoundedFloatImpl.valueOf(MIN - 1));
		log.info(BoundedFloatImpl.valueOf(MIN + 1));
		log.info(BoundedFloatImpl.valueOf(MAX + 1));
		log.info(BoundedFloatImpl.valueOf(MAX - 1));
	}

	/**
	 * Returns a Float object for value of `f`, after checking boundary
	 * constraints
	 *
	 * XXX FIXME How to make that funky beast working as template such that
	 * inheriting classes do not need to override it explicitly.
	 *
	 * @param f
	 * @return a Float object for value of `f`
	 */
	public static Float valueOf(float f) {
		return new BoundedFloatImpl(f, MIN, MAX).floatValue();
	}

	/** {@inheritDoc} */
	public String toString() {
		return String.valueOf(value);
	}

	/** {@inheritDoc} */
	public double doubleValue() {
		return (double) value;
	}

	/** {@inheritDoc} */
	public float floatValue() {
		return value;
	}

	/** {@inheritDoc} */
	public int intValue() {
		return (int) value;
	}

	/** {@inheritDoc} */
	public long longValue() {
		return (long) value;
	}

	/** {@inheritDoc} */
	public int compareTo(BoundedFloatImpl o) {
		return Float.compare(this.value, o.value);
	}

	/**
	 * XXX TODO If we leave this out, each `BoundedIntegerImpl`- and
	 * derivative-classified object should have a normal Object `hashCode()`
	 * value as a result
	 */
	/** {@inheritDoc} */
	public int hashCode() {
		return Float.floatToIntBits(value);
	}
}
