/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.types.impl;

import static org.junit.Assert.assertTrue;

import java.util.Collection;

import repast.simphony.valueLayer.GridValueLayer;
import abmland.common.LayerEvent;
import abmland.common.LayerListener;
import abmland.common.LayerEvent.Type;
import abmland.data.DatatypeBoundaries;
import abmland.types.INumber;

public class BoundedFloatValueLayer<F extends INumber<Float>> extends
		AbstractAllCells<F> {

	private float LOWER;
	private float UPPER;

	private void handleBoundaries(Float value) throws AssertionError {
		try {
			assertTrue(
					value
							+ " out of borders.",
					(value >= LOWER) && (value <= UPPER));
		} catch (AssertionError e) {
			layerEventOccurred(new LayerEvent<F>(this, value,
					Type.OBJECT_INVALID));
			e.printStackTrace();
		}
	}

	/**
	 * @tag usage.protocol -rating="4" : Take care! defaultValue has to be the same as that used to construct the
	 * GridValueLayer object!
	 *
	 * @param layer
	 * @param bounded
	 * @param defaultValue
	 */
	public BoundedFloatValueLayer(GridValueLayer layer, F bounded,
			Float defaultValue) {
		super(layer);
		this.LOWER = bounded.MIN();
		this.UPPER = bounded.MAX();
		handleBoundaries(defaultValue);
	}

	/**
	 * @tag usage.protocol -rating="4" : Take care! defaultValue has to be the same as that used to construct the
	 * GridValueLayer object!
	 *
	 * @param layer
	 * @param min
	 * @param max
	 * @param defaultValue
	 */
	public BoundedFloatValueLayer(GridValueLayer layer, float min, float max,
			Float defaultValue) {
		super(layer);
		this.LOWER = min;
		this.UPPER = max;
		handleBoundaries(defaultValue);
	}

	/**
	 * @tag usage.protocol -rating="4" : Take care! defaultValue has to be the same as that used to construct the
	 * GridValueLayer object!
	 *
	 * @param layer
	 * @param min
	 * @param max
	 * @param defaultValue
	 */
	public BoundedFloatValueLayer(GridValueLayer layer, DatatypeBoundaries min,
			DatatypeBoundaries max, Float defaultValue) {
		super(layer);
		this.LOWER = min.value().floatValue();
		this.UPPER = max.value().floatValue();
		handleBoundaries(defaultValue);
	}

	/**
	 * @tag usage.general -rating="4" : We will raise the layer event `Type.OBJECT_INVALID` and throw an
	 * exception, if a wrong-bounded value would have been stored.
	 */
	/** {@inheritDoc} */
	public void set(double value, int... coordinate) {
		float floatvalue = new Float(value);
		try {
			assertTrue(value + " out of borders.", (floatvalue >= LOWER)
					&& (floatvalue <= UPPER));
			/**
			 * Only store if no [exception thrown until now] boundaries broken
			 */
			super.set(value, coordinate);
			layerEventOccurred(new LayerEvent<F>(this, value,
					LayerEvent.OBJECT_ADDED));

		} catch (AssertionError e) {
			layerEventOccurred(new LayerEvent<F>(this, value,
					Type.OBJECT_INVALID));
			e.printStackTrace();
		}
	}

	/** {@inheritDoc} */
	public void layerEventOccurred(LayerEvent<F> evt) {
		Collection<LayerListener<F>> listeners = getLayerListeners();
		for (LayerListener<F> layerListener : listeners) {
			layerListener.layerEventOccurred(evt);
		}
	}

	/** {@inheritDoc} */
	public void set(F value, int... coordinate) {
		set(value.doubleValue(), coordinate);
	}
}
