/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.types.impl;

import static abmland.types.impl.HouseholdNumber.getInstance;
import static abmland.types.impl.TypeUtil.guardAssignment;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

import abmland.types.AbstractBoundedDouble;
import abmland.types.BoundedDouble;
import abmland.types.IHouseholdNumber;
import static abmland.common.LogExceptionHandler.handleException;

/**
 * Take this as a model class for the new sub-hierarchy of numbers, where you
 * may look for the principal ideas first comings. Add new ones here first.
 *
 * @author dkahlenb
 *
 */
@BoundedDouble(MIN = 0, MAX = 1e5, NODATA = Double.NaN)
public class HouseholdNumber extends AbstractBoundedDouble implements
		IHouseholdNumber {

	private double value = HouseholdNumber.NODATA;

	// if PERIPHERIC is true, this value relates to a non-CS area grid point and
	// Double.NaN is
	// allowed as internal representation, if not, an exception should be
	// thrown, marking
	// the Double.NaN to be result of failure (division by zero etc.).
	private boolean PERIPHERIC = false;

	private final static Class<BoundedDouble> inherited_ = BoundedDouble.class;
	private final static Class<HouseholdNumber> this_ = HouseholdNumber.class;

	public final static Double MIN = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MIN()
			: null;

	public final static Double MAX = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MAX()
			: null;

	public final static Double NODATA = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).NODATA()
			: null;

	/**
	 * {@FIXME p:Paper We could store any given NaN as -9999 or what is the
	 * NODATA representation here inside the constructor tail, but I'm advising
	 * against doing so, at least currently}
	 *
	 * @param value
	 * @param fromPeriphery
	 *            If there is no argument given here, fromPeriphery will though
	 *            have size zero but isn't null.
	 */
	public HouseholdNumber(Number value, boolean... fromPeriphery) throws RuntimeException {
		// in this first case, NaN is stored
		if (fromPeriphery.length > 0 && fromPeriphery[0]) {
			this.PERIPHERIC = true;
//			try {
				try {
					this.value = guardAssignment(value.doubleValue(), MIN, MAX,
							NODATA);
				} catch (Exception e) {
					throw new RuntimeException(e);
				}
//			} catch (Exception e) {
//				handleException(HouseholdNumber.class, e);
//			}
		} else if ((fromPeriphery.length > 0 && !fromPeriphery[0])
				|| fromPeriphery.length <= 0) {
			// Nina said inside the CS area polygon there never will be NODATA
			// points, which is according to the ASCII-Raster-Specification I
			// guess... So we let guardAssignment handle the exception to be
			// thrown if a value of NaN (NODATA marker) is given by calling it
			// with MAX as NODATA, so it detects NaN != MAX. If that's no
			// reliable solution the more explicit approach handling also aother
			// values than NaN as NODATA representations is below inside
			// /**explicit*/
			// If the given value is stored at no peripheric point, we don't
			// need a NODATA definition value, simply give duplicate MAX...
			this.PERIPHERIC = false;
//			try {
				try {
					this.value = guardAssignment(value.doubleValue(), MIN, MAX, MAX);
				} catch (Exception e) {
					throw new RuntimeException(e);
				}
//			} catch (Exception e) {
//				handleException(HouseholdNumber.class, e);
//			}
			// /**explicit*/
			// if (value == HouseholdNumber.NODATA) {
			// // throw exception
			// handleException(
			// HouseholdNumber.class,
			// new IllegalArgumentException(
			// value
			// +
			// " given. This may result from illegal operation (with undefined result). For marking peripheric data give boolean parameter set to true."));
			// // - OR -
			// // but that's very implicit, observe the MAX given instead of
			// // NODATA, so the guardAssignment method will throw the
			// // exception here, because value is NaN, except NODATA is
			// // defined as MIN <= NODATA <= MAX which obviously would make no
			// // sense...
			// // this.value = guardAssignment(value.doubleValue(), MIN, MAX,
			// // MAX);
			// }
			// // assuming the exception stopped the runtime system if former
			// // occurred, we don't need else branch here
			// // If it is no peripheric point we don't need no NODATA
			// this.value = guardAssignment(value.doubleValue(), MIN, MAX, MAX);
			// /**explicit*/
		}

		// both cases, fromPeriphery[0]==false and fromPeriphery not given are
		// handled the same
	}

	public HouseholdNumber(boolean... fromPeriphery) {
		this.value = NODATA;
		// in this first case, NaN is stored
		if (fromPeriphery.length > 0 && fromPeriphery[0]) {
			this.PERIPHERIC = true;
			this.value = NODATA;
		} else if ((fromPeriphery.length > 0 && !fromPeriphery[0])
				|| 0 >= fromPeriphery.length) {
			this.PERIPHERIC = false;
			this.value = MIN;
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see abmland.types.impl.IHouseholdNumber#MAX()
	 */
	public double MAX() {
		return HouseholdNumber.MAX;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see abmland.types.impl.IHouseholdNumber#MIN()
	 */
	public double MIN() {
		return HouseholdNumber.MIN;
	}

	@Override
	public double NODATA() {
		return HouseholdNumber.NODATA;
	}

	/** {@inheritDoc} */
	public double value() {
		return value;
	}

	/**
	 * {@inheritDoc}
	 *
	 * Need bound checking here, means an instance created, so this cannot be in
	 * the abstract super class.
	 *
	 * @param i
	 * @return ((an Double value for the double??)) stored by this instance of
	 *         NewPriceIndex
	 *
	 *         TODO Should we return a BoundedDouble here rather or even an
	 *         Double - what is the semantics of valueOf e. g. in Double?
	 *
	 *         XXX (1) From a technical view podouble it is more attractive to
	 *         have a usage of this type like (using static imports though):
	 *         valueOf(1) + valueOf(2) than like: valueOf(1).value() +
	 *         valueOf(2).value(), the latter which being the case if we return
	 *         NewHouseholdNumber instead of Double here.
	 *
	 *         XXX (2) From a scientific view podouble there may come
	 *         consequences when doubleroducing such a generalization. maybe
	 *         those a ignore-able if we really hinder each other usage than the
	 *         boundary checking with that classes.
	 */
	public final static Double valueOf(double i, boolean... fromPeriphery) throws RuntimeException {
		return new HouseholdNumber(i, fromPeriphery).value();
	}

	public static HouseholdNumber getInstance(Number n,
			boolean... fromPeriphery) throws RuntimeException {
		return new HouseholdNumber(n.doubleValue(), fromPeriphery);
	}

	// TODO Should we rather leave that out from the NewAbstractBounded... and
	// change
	// the parameter type to NewHouseholdNumber here, having to add own
	// implementation in the methods tail?
	/** {@inheritDoc} */
	public int compareTo(AbstractBoundedDouble another) {
		return super.compareTo(another);
	}

	/** {@inheritDoc} */
	public boolean equals(Object obj) {
		return (obj instanceof HouseholdNumber)
				&& value() == ((HouseholdNumber) obj).value();
	}
}
