/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.types.impl;

import static abmland.types.impl.TypeUtil.guardAssignment;
import abmland.types.AbstractBoundedDouble;
import abmland.types.BoundedDouble;
import abmland.types.IPressure;

/**
 * @author dkahlenb
 *
 */
@BoundedDouble(MIN = -1.0d, MAX = 1.0d, NODATA = Double.NaN)
public final class Pressure extends AbstractBoundedDouble implements IPressure {

	private double value = Pressure.NODATA;

	// if PERIPHERIC is true, this value relates to a non-CS area grid point and
	// Double.NaN is
	// allowed as internal representation, if not, an exception should be
	// thrown, marking
	// the Double.NaN to be result of failure (division by zero etc.).
	private boolean PERIPHERIC = false;

	private static Class<BoundedDouble> inherited_ = BoundedDouble.class;
	private static Class<Pressure> this_ = Pressure.class;

	public static final Double MIN = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MIN()
			: null;

	public static final Double MAX = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MAX()
			: null;

	public final static Double NODATA = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).NODATA()
			: null;

	public Pressure(Number value, boolean... fromPeriphery)
			throws RuntimeException {
		// in this first case, NaN is stored
		if (fromPeriphery.length > 0 && fromPeriphery[0]) {
			this.PERIPHERIC = true;
			try {
				this.value = guardAssignment(value.doubleValue(), MIN, MAX,
						NODATA);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		} else if ((fromPeriphery.length > 0 && !fromPeriphery[0])
				|| fromPeriphery.length <= 0) {
			this.PERIPHERIC = false;
			try {
				this.value = guardAssignment(value.doubleValue(), MIN, MAX, MAX);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}

	public Pressure(boolean... fromPeriphery) {
		this.value = NODATA;
		// in this first case, NaN is stored
		if (fromPeriphery.length > 0 && fromPeriphery[0]) {
			this.PERIPHERIC = true;
			this.value = NODATA;
		} else if ((fromPeriphery.length > 0 && !fromPeriphery[0])
				|| 0 >= fromPeriphery.length) {
			this.PERIPHERIC = false;
			this.value = MIN;
		}
	}

	/** {@inheritedDoc} */
	public final double MAX() {
		return MAX;
	}

	/** {@inheritedDoc} */
	public final double MIN() {
		return MIN;
	}

	@Override
	public double NODATA() {
		return Pressure.NODATA;
	}

	/** {@inheritedDoc} */
	public final double value() {
		return value;
	}

	public final static Double valueOf(double f, boolean... fromPeriphery)
			throws RuntimeException {
		return new Pressure(f, fromPeriphery).value();
	}

	// TODO Return a BoundedFloat here rather?
	public static Pressure getInstance(Number n, boolean... fromPeriphery)
			throws RuntimeException {
		return new Pressure(n.doubleValue(), fromPeriphery);
	}

	/** {@inheritDoc} */
	public int compareTo(AbstractBoundedDouble another) {
		return super.compareTo(another);
	}

	/** {@inheritDoc} */
	public boolean equals(Object obj) {
		return (obj instanceof Pressure) && value() == ((Pressure) obj).value();
	}
}
