/**
 * Copyright (C) 2010 [[XYZ]], Helmholtz Centre for Environmental Research GmbH - UFZ <abmland@ufz.de>
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package abmland.types.impl;

import static abmland.types.impl.TypeUtil.guardAssignment;
import abmland.types.AbstractBoundedDouble;
import abmland.types.AbstractBoundedInteger;
import abmland.types.BoundedDouble;
import abmland.types.BoundedInteger;
import abmland.types.ISpace;

/**
 * For pragmatic reasons we decide to make this a Double not an Integer, needs
 * NaN, which isn't available with Integer, casted to 0 only, which is
 * conflicting with our range def.
 *
 * @author dkahlenb
 *
 */
@BoundedDouble(MIN = 0, MAX = 1e4, NODATA = Double.NaN)
public class Space extends AbstractBoundedDouble implements ISpace {

	private double value = Space.NODATA;

	// if PERIPHERIC is true, this value relates to a non-CS area grid point and
	// Double.NaN is
	// allowed as internal representation, if not, an exception should be
	// thrown, marking
	// the Double.NaN to be result of failure (division by zero etc.).
	private boolean PERIPHERIC = false;

	private final static Class<BoundedDouble> inherited_ = BoundedDouble.class;
	private final static Class<Space> this_ = Space.class;

	public final static Double MIN = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MIN()
			: null;

	public final static Double MAX = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).MAX()
			: null;

	public final static Double NODATA = this_.isAnnotationPresent(inherited_) ? this_
			.getAnnotation(inherited_).NODATA()
			: null;

	public Space(Number value, boolean... fromPeriphery)
			throws RuntimeException {
		// in this first case, NaN is stored
		if (fromPeriphery.length > 0 && fromPeriphery[0]) {
			this.PERIPHERIC = true;
			try {
				this.value = guardAssignment(value.doubleValue(), MIN, MAX,
						NODATA);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		} else if ((fromPeriphery.length > 0 && !fromPeriphery[0])
				|| fromPeriphery.length <= 0) {
			this.PERIPHERIC = false;
			try {
				this.value = guardAssignment(value.doubleValue(), MIN, MAX, MAX);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}

	public Space(boolean... fromPeriphery) {
		this.value = NODATA;
		// in this first case, NaN is stored
		if (fromPeriphery.length > 0 && fromPeriphery[0]) {
			this.PERIPHERIC = true;
			this.value = NODATA;
		} else if ((fromPeriphery.length > 0 && !fromPeriphery[0])
				|| 0 >= fromPeriphery.length) {
			this.PERIPHERIC = false;
			this.value = MIN;
		}
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see abmland.types.impl.ISpace#MAX()
	 */
	public double MAX() {
		return Space.MAX;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see abmland.types.impl.ISpace#MIN()
	 */
	public double MIN() {
		return Space.MIN;
	}

	@Override
	public double NODATA() {
		return Space.NODATA;
	}

	/** {@inheritDoc} */
	public double value() {
		return value;
	}

	/**
	 * {@inheritDoc}
	 *
	 * Need bound checking here, means an instance created, so this cannot be in
	 * the abstract super class.
	 *
	 * @param i
	 * @return ((an Integer value for the int??)) stored by this instance of
	 *         NewPriceIndex
	 *
	 *         TODO Should we return a BoundedInteger here rather or even an
	 *         Integer - what is the semantics of valueOf e. g. in Integer?
	 *
	 *         XXX (1) From a technical view point it is more attractive to have
	 *         a usage of this type like (using static imports though):
	 *         valueOf(1) + valueOf(2) than like: valueOf(1).value() +
	 *         valueOf(2).value(), the latter which being the case if we return
	 *         NewSpace instead of Integer here.
	 *
	 *         XXX (2) From a scientific view point there may come consequences
	 *         when introducing such a generalization. maybe those a ignore-able
	 *         if we really hinder each other usage than the boundary checking
	 *         with that classes.
	 */
	public final static Double valueOf(double i, boolean... fromPeriphery)
			throws RuntimeException {
		return new Space(i, fromPeriphery).value();
	}

	/**
	 * TODO Return a BoundedInteger here rather? But the term-intrinsic
	 * semantics rather point to an instance of exactly this class here.
	 */
	public static Space getInstance(Number n, boolean... fromPeriphery)
			throws RuntimeException {
		return new Space(n.doubleValue(), fromPeriphery);
	}

	// TODO Should we rather leave that out from the NewAbstractBounded... and
	// change
	// the parameter type to NewSpace here, having to add own
	// implementation in the methods tail?
	/** {@inheritDoc} */
	public int compareTo(AbstractBoundedDouble another) {
		return super.compareTo(another);
	}

	/** {@inheritDoc} */
	public boolean equals(Object obj) {
		return (obj instanceof Space) && value() == ((Space) obj).value();
	}
}
